/* SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause */
/*
 * Copyright 2021 Amazon.com, Inc. or its affiliates. All rights reserved.
 */

#ifndef _EFA_IO_REGS_H_
#define _EFA_IO_REGS_H_

/* cq_db register */
#define EFA_IO_REGS_CQ_DB_CONSUMER_INDEX_MASK               0xffff
#define EFA_IO_REGS_CQ_DB_CMD_SN_MASK                       0x60000000
#define EFA_IO_REGS_CQ_DB_ARM_MASK                          0x80000000

#endif /* _EFA_IO_REGS_H_ */
